import inspect
from functools import wraps
from typing import Callable, Awaitable, Any

from bolinette import blnt
from bolinette.testing import TestClient, Bolitest
from bolinette.utils.functions import async_invoke


def bolitest(*, before: Callable[[Any], Awaitable[None]] = None,
             after: Callable[[Any], Awaitable[None]] = None):
    def wrapper(func):
        @wraps(func)
        async def inner(client: TestClient):
            await client.context.db.drop_all()
            await client.context.db.create_all()
            if before is not None:
                await async_invoke(before, context=client.context, mock=client.mock)
            await client.context.db.close_transaction()
            await func(client=client)
            if after is not None:
                await async_invoke(after, context=client.context, mock=client.mock)
            await client.context.db.drop_all()
        blnt.cache.test_funcs.append(Bolitest(inner, inspect.getfile(func)))
        return inner
    return wrapper



from bolinette.core.repositories.repository import Repository
from bolinette.core.repositories.relational import RelationalRepository
from bolinette.core.repositories.collection import CollectionRepository



from bolinette import blnt, core
from bolinette.exceptions import APIErrors, ParamConflictError
from bolinette.utils.functions import getattr_, setattr_


class Repository:
    def __init__(self, name: str, model: 'core.Model', context: 'blnt.BolinetteContext'):
        self.name = name
        self.model = model
        self.context = context

    async def get_all(self, pagination=None, order_by=None):
        raise NotImplementedError()

    async def get(self, identifier):
        raise NotImplementedError()

    async def get_by(self, key, value):
        raise NotImplementedError()

    async def get_first_by(self, key, value):
        raise NotImplementedError()

    async def get_by_criteria(self, criteria):
        raise NotImplementedError()

    async def create(self, values):
        raise NotImplementedError()

    async def update(self, entity, values):
        raise NotImplementedError()

    async def patch(self, entity, values):
        raise NotImplementedError()

    async def delete(self, entity):
        raise NotImplementedError()

    async def _validate_model(self, values: dict):
        api_errors = APIErrors()
        for column in self.model.__props__.get_columns().values():
            key = column.name
            if column.primary_key:
                continue
            if key in values:
                value = values.get(key)
                if column.unique:
                    if await self.get_first_by(column.name, value) is not None:
                        api_errors.append(ParamConflictError(key, value))
        if api_errors:
            raise api_errors
        return values

    async def _map_model(self, entity, values, patch=False):
        api_errors = APIErrors()
        for _, column in self.model.__props__.get_columns().items():
            key = column.name
            if column.primary_key or (key not in values and patch):
                continue
            original = getattr_(entity, key, None)
            new = values.get(key, None)
            if original == new:
                continue
            if column.unique and new is not None:
                if await self.get_first_by(column.name, new) is not None:
                    api_errors.append(ParamConflictError(key, new))
            setattr_(entity, key, new)
        if api_errors:
            raise api_errors



from bolinette.blnt.database.pagination import Pagination
from bolinette.blnt.database.engine import DatabaseEngine
from bolinette.blnt.database.relational import RelationalDatabase
from bolinette.blnt.database.collection import CollectionDatabase
from bolinette.blnt.database.manager import DatabaseManager



from bolinette import blnt, core
from bolinette.blnt.database import Pagination
from bolinette.core.repositories import Repository


class RelationalRepository(Repository):
    def __init__(self, name: str, model: 'core.Model', context: 'blnt.BolinetteContext'):
        super().__init__(name, model, context)
        self.database: 'blnt.database.RelationalDatabase' = context.db[model.__blnt__.database]
        self.table = context.table(name)

    def __repr__(self):
        return f'<Repository {self.name}>'

    @property
    def query(self):
        return self.database.session.query(self.table)

    async def get_all(self, pagination=None, order_by=None):
        if order_by is None:
            order_by = []
        query = self.query
        if len(order_by) > 0:
            query = await self._build_order_by(query, order_by)
        if pagination is not None:
            return self._paginate(query, pagination)
        return query.all()

    async def get(self, identifier):
        return self.query.get(identifier)

    async def get_by(self, key, value):
        return self.query.filter_by(**{key: value}).all()

    async def get_first_by(self, key, value):
        return self.query.filter_by(**{key: value}).first()

    async def get_by_criteria(self, criteria):
        return self.query.filter(criteria).all()

    async def create(self, values):
        filtered = await self._validate_model(values)
        entity = self.table(**filtered)
        self.database.session.add(entity)
        return entity

    async def update(self, entity, values):
        await self._map_model(entity, values)
        return entity

    async def patch(self, entity, values):
        await self._map_model(entity, values, patch=True)
        return entity

    async def delete(self, entity):
        self.database.session.delete(entity)
        return entity

    async def _build_order_by(self, query, params):
        order_by_query = []
        for col_name, way in params:
            if hasattr(self.table, col_name):
                column = getattr(self.table, col_name)
                if way:
                    order_by_query.append(column)
                else:
                    order_by_query.append(core.functions.desc(column))
        return query.order_by(*order_by_query)

    @staticmethod
    def _paginate(query, pagination):
        page = pagination['page']
        per_page = pagination['per_page']
        total = query.count()
        items = query.offset(page * per_page).limit(per_page).all()
        return Pagination(items, page, per_page, total)



from bson import ObjectId

from bolinette import blnt, core
from bolinette.core.repositories import Repository
from bolinette.exceptions import InternalError


class CollectionRepository(Repository):
    def __init__(self, name: str, model: 'core.Model', context: 'blnt.BolinetteContext'):
        super().__init__(name, model, context)
        self.database: 'blnt.database.CollectionDatabase' = context.db[model.__blnt__.database]

    @property
    def collection(self):
        return self.database.db[self.name]

    async def get_all(self, pagination=None, order_by=None):
        return self.collection.find()

    async def get(self, identifier):
        return self.collection.find_one({'_id': ObjectId(identifier)})

    async def get_by(self, key, value):
        if key == '_id':
            value = ObjectId(value)
        return self.collection.find({key: value})

    async def get_first_by(self, key, value):
        if key == '_id':
            value = ObjectId(value)
        return self.collection.find_one({key: value})

    async def get_by_criteria(self, criteria):
        raise InternalError('internal.feature.not_supported')

    async def create(self, values):
        filtered = await self._validate_model(values)
        result = self.collection.insert_one(filtered)
        return await self.get(result.inserted_id)

    async def update(self, entity, values):
        _id = entity['_id']
        del entity['_id']
        await self._map_model(entity, values)
        self.collection.update_one(
            {'_id': _id},
            {'$set': entity},
            upsert=False
        )
        return await self.get(_id)

    async def patch(self, entity, values):
        _id = entity['_id']
        del entity['_id']
        await self._map_model(entity, values, patch=True)
        self.collection.update_one(
            {'_id': _id},
            {'$set': entity},
            upsert=False
        )
        return await self.get(_id)

    async def delete(self, entity):
        self.collection.delete_one({'_id': entity['_id']})
        return entity



class DatabaseEngine:
    def __init__(self, relational: bool):
        self.relational = relational

    async def open_transaction(self):
        raise NotImplementedError()

    async def close_transaction(self):
        raise NotImplementedError()

    async def rollback_transaction(self):
        raise NotImplementedError()

    async def create_all(self):
        raise NotImplementedError()

    async def drop_all(self):
        raise NotImplementedError()



from typing import Dict, Any, Callable, Awaitable

from aiohttp import web as aio_web
from aiohttp.web_request import Request
from aiohttp.web_response import Response

from bolinette import web, blnt
from bolinette.blnt.database import Pagination
from bolinette.decorators import middleware
from bolinette.exceptions import BadRequestError
from bolinette.utils.serializing import deserialize, serialize


@middleware('blnt_payload', priority=6, auto_load=False, loadable=False)
class PayloadMiddleware(web.InternalMiddleware):
    async def handle(self, request, params, next_func):
        try:
            payload = await deserialize(request)
        except Exception:
            raise BadRequestError('global.payload.unserializable')
        if 'model' in self.options:
            payload = self.context.validator.validate_payload(self.options['model'], self.options['key'],
                                                              payload, self.options.get('patch'))
            await self.context.validator.link_foreign_entities(self.options['model'], self.options['key'], payload)
        params['payload'] = payload
        return await next_func(request, params)


@middleware('blnt_headers', priority=0, auto_load=True, loadable=False)
class HeadersMiddleware(web.InternalMiddleware):
    async def handle(self, request: Request, params: Dict[str, Any],
                     next_func: Callable[[Request, Dict[str, Any]], Awaitable[Response]]):
        params['headers'] = {}
        for key in request.headers:
            params['headers'][key] = request.headers[key]
        return await next_func(request, params)


@middleware('blnt_response', priority=4, auto_load=False, loadable=False)
class ResponseMiddleware(web.InternalMiddleware):
    async def handle(self, request, params, next_func):
        async with blnt.Transaction(self.context):
            resp = await next_func(request, params)
        if resp is None:
            return aio_web.Response(status=204)
        if isinstance(resp, aio_web.Response):
            return resp
        if isinstance(resp, str):
            return aio_web.Response(text=resp, status=200, content_type='text/plain')
        if not isinstance(resp, web.APIResponse):
            return aio_web.Response(text='global.response.unserializable', status=500, content_type='text/plain')

        content = resp.content

        if content.get('data') is not None and isinstance(content['data'], Pagination):
            content['pagination'] = {
                'page': content['data'].page,
                'per_page': content['data'].per_page,
                'total': content['data'].total,
            }
            content['data'] = content['data'].items

        if 'model' in self.options:
            ret_def = self.context.mapper.response(self.options['model'], self.options['key'])
            if content.get('data') is not None:
                content['data'] = self.context.mapper.marshall(ret_def, content['data'],
                                                               skip_none=self.options.get('skip_none', False),
                                                               as_list=self.options.get('as_list', False))

        serialized, mime = serialize(content, 'application/json')

        web_response = aio_web.Response(text=serialized, status=resp.code, content_type=mime)
        for cookie in resp.cookies:
            if not cookie.delete:
                web_response.set_cookie(cookie.name, cookie.value,
                                        expires=cookie.expires.strftime('%a, %d %b %Y %H:%M:%S GMT'),
                                        path=cookie.path, httponly=cookie.http_only)
            else:
                web_response.del_cookie(cookie.name, path=cookie.path)

        return web_response



from bolinette.blnt import database
from bolinette.blnt.logging import Logger
from bolinette.blnt.cache import cache
from bolinette.blnt.environment import init, Environment
from bolinette.blnt.jwt import JWT
from bolinette.blnt.validation import Validator
from bolinette.blnt.context import BolinetteContext
from bolinette.blnt.transaction import Transaction
